@@ -15,8 +15,10 @@ module Codec.Encryption.OpenPGP.Types where
 
 import GHC.Generics (Generic)
 
+import Control.Applicative ((<$>), (<|>))
 import Control.Arrow ((***))
 import Control.Lens (makeLenses)
+import Control.Monad (mzero)
 import Control.Newtype (Newtype(..))
 import qualified Crypto.PubKey.RSA as RSA
 import qualified Crypto.PubKey.DSA as DSA
@@ -35,6 +37,7 @@ import Data.List (unfoldr)
 import Data.List.NonEmpty (NonEmpty)
 import qualified Data.List.NonEmpty as NE
 import Data.List.Split (chunksOf)
+import Data.Maybe (fromMaybe)
 import Data.Monoid ((<>), Monoid, mempty)
 import Data.Ord (comparing)
 import Data.Set (Set)
@@ -45,7 +48,7 @@ import Data.Time.Clock.POSIX (posixSecondsToUTCTime)
 import Data.Time.Format (formatTime)
 import Data.Typeable (Typeable)
 import Data.Word (Word8, Word16, Word32)
-import Network.URI (URI(..), uriToString)
+import Network.URI (URI(..), uriToString, nullURI, parseURI)
 import Numeric (readHex, showHex)
 import System.Locale (defaultTimeLocale)
 import Text.PrettyPrint.Free (Pretty(..), (<+>), char, hsep, punctuate, space, text, tupled)
@@ -57,8 +60,6 @@ type AlmostPublicDomainRegex = ByteString
 type Revocability = Bool
 type RevocationReason = Text
 type KeyServer = ByteString
-type NotationName = ByteString
-type NotationValue = ByteString
 type SignatureHash = ByteString
 type PacketVersion = Word8
 type V3Expiration = Word16
@@ -85,7 +86,7 @@ instance Eq SymmetricAlgorithm where
     (==) a b = fromFVal a == fromFVal b
 
 instance Ord SymmetricAlgorithm where
-    compare a b = fromFVal a `compare` fromFVal b
+    compare = comparing fromFVal
 
 instance FutureVal SymmetricAlgorithm where
     fromFVal Plaintext = 0
@@ -130,6 +131,7 @@ instance Pretty SymmetricAlgorithm where
     pretty (OtherSA sa) = text "unknown symmetric algorithm" <+> (text . show) sa
 
 instance A.ToJSON SymmetricAlgorithm
+instance A.FromJSON SymmetricAlgorithm
 
 data NotationFlag = HumanReadable
                   | OtherNF Word8 -- FIXME: this should be constrained to 4 bits?
@@ -139,7 +141,7 @@ instance Eq NotationFlag where
     (==) a b = fromFFlag a == fromFFlag b
 
 instance Ord NotationFlag where
-    compare a b = fromFFlag a `compare` fromFFlag b
+    compare = comparing fromFFlag
 
 instance FutureFlag NotationFlag where
     fromFFlag HumanReadable = 0
@@ -155,6 +157,7 @@ instance Pretty NotationFlag where
     pretty (OtherNF o) = text "unknown notation flag type" <+> pretty o
 
 instance A.ToJSON NotationFlag
+instance A.FromJSON NotationFlag
 
 data SigSubPacket = SigSubPacket {
     _sspCriticality :: Bool
@@ -167,6 +170,7 @@ instance Pretty SigSubPacket where
 instance Hashable SigSubPacket
 
 instance A.ToJSON SigSubPacket
+instance A.FromJSON SigSubPacket
 
 newtype ThirtyTwoBitTimeStamp = ThirtyTwoBitTimeStamp {unThirtyTwoBitTimeStamp :: Word32}
     deriving (Bounded, Data, Enum, Eq, Generic, Hashable, Integral, Num, Ord, Real, Show, Typeable)
@@ -179,6 +183,7 @@ instance Pretty ThirtyTwoBitTimeStamp where
     pretty = text . formatTime defaultTimeLocale "%Y%m%d-%H%M%S" . posixSecondsToUTCTime . realToFrac
 
 instance A.ToJSON ThirtyTwoBitTimeStamp
+instance A.FromJSON ThirtyTwoBitTimeStamp
 
 newtype ThirtyTwoBitDuration = ThirtyTwoBitDuration {unThirtyTwoBitDuration :: Word32}
     deriving (Bounded, Data, Enum, Eq, Generic, Hashable, Integral, Num, Ord, Real, Show, Typeable)
@@ -191,6 +196,7 @@ instance Pretty ThirtyTwoBitDuration where
     pretty = text . concat . unfoldr durU . unpack
 
 instance A.ToJSON ThirtyTwoBitDuration
+instance A.FromJSON ThirtyTwoBitDuration
 
 durU :: (Integral a, Show a) => a -> Maybe (String, a)
 durU x
@@ -214,7 +220,39 @@ instance Pretty URL where
     pretty = pretty . (\uri -> uriToString id uri "") . unpack
 
 instance A.ToJSON URL where
-    toJSON = A.toJSON . (\uri -> uriToString id uri "") . unpack
+    toJSON u = object [T.pack "uri" .= (\uri -> uriToString id uri "") (unpack u)]
+instance A.FromJSON URL where
+    parseJSON (A.Object v) = URL . fromMaybe nullURI . parseURI <$>
+                                      v A..: T.pack "uri"
+    parseJSON _            = mzero
+
+newtype NotationName = NotationName {unNotationName :: ByteString}
+    deriving (Data, Eq, Generic, Hashable, Ord, Pretty, Show, Typeable)
+
+instance Newtype NotationName ByteString where
+    pack = NotationName
+    unpack (NotationName nn) = nn
+
+instance A.ToJSON NotationName where
+    toJSON nn = object [T.pack "notationname" .= show (unpack nn)]
+instance A.FromJSON NotationName where
+    parseJSON (A.Object v) = NotationName . read <$>
+                                      v A..: T.pack "notationname"
+    parseJSON _            = mzero
+
+newtype NotationValue = NotationValue {unNotationValue :: ByteString}
+    deriving (Data, Eq, Generic, Hashable, Ord, Pretty, Show, Typeable)
+
+instance Newtype NotationValue ByteString where
+    pack = NotationValue
+    unpack (NotationValue nv) = nv
+
+instance A.ToJSON NotationValue where
+    toJSON nv = object [T.pack "notationvalue" .= show (unpack nv)]
+instance A.FromJSON NotationValue where
+    parseJSON (A.Object v) = NotationValue . read <$>
+                                      v A..: T.pack "notationvalue"
+    parseJSON _            = mzero
 
 data SigSubPacketPayload = SigCreationTime ThirtyTwoBitTimeStamp
                   | SigExpirationTime ThirtyTwoBitDuration
@@ -273,31 +311,48 @@ instance Pretty SigSubPacketPayload where
     pretty (OtherSigSub t bs) = text "unknown signature subpacket type" <+> pretty t <+> pretty bs
 
 instance A.ToJSON SigSubPacketPayload where
-    toJSON (SigCreationTime ts) = A.toJSON ts
-    toJSON (SigExpirationTime d) = A.toJSON d
-    toJSON (ExportableCertification e) = A.toJSON e
-    toJSON (TrustSignature tl ta) = A.toJSON (tl, ta)
-    toJSON (RegularExpression apdre) = A.toJSON (BL.unpack apdre)
-    toJSON (Revocable r) = A.toJSON r
-    toJSON (KeyExpirationTime d) = A.toJSON d
-    toJSON (PreferredSymmetricAlgorithms sas) = A.toJSON sas
-    toJSON (RevocationKey rcs pka tof) = A.toJSON (rcs, pka, tof)
-    toJSON (Issuer eoki) = A.toJSON eoki
-    toJSON (NotationData nfs nn nv) = A.toJSON (nfs, BL.unpack nn, BL.unpack nv)
-    toJSON (PreferredHashAlgorithms phas) = A.toJSON phas
-    toJSON (PreferredCompressionAlgorithms pcas) = A.toJSON pcas
-    toJSON (KeyServerPreferences kspfs) = A.toJSON kspfs
-    toJSON (PreferredKeyServer ks) = A.toJSON (show ks)
-    toJSON (PrimaryUserId p) = A.toJSON p
-    toJSON (PolicyURL u) = A.toJSON u
-    toJSON (KeyFlags kfs) = A.toJSON kfs
-    toJSON (SignersUserId u) = A.toJSON u
-    toJSON (ReasonForRevocation rc rr) = A.toJSON (rc, rr)
-    toJSON (Features ffs) = A.toJSON ffs
-    toJSON (SignatureTarget pka ha sh) = A.toJSON (pka, ha, BL.unpack sh)
-    toJSON (EmbeddedSignature sp) = A.toJSON sp
-    toJSON (UserDefinedSigSub t bs) = A.toJSON (t, BL.unpack bs)
-    toJSON (OtherSigSub t bs) = A.toJSON (t, BL.unpack bs)
+    toJSON (SigCreationTime ts) = object [T.pack "sigCreationTime" .= ts]
+    toJSON (SigExpirationTime d) = object [T.pack "sigExpirationTime" .= d]
+    toJSON (ExportableCertification e) = object [T.pack "exportableCertification" .= e]
+    toJSON (TrustSignature tl ta) = object [T.pack "trustSignature" .= (tl, ta)]
+    toJSON (RegularExpression apdre) = object [T.pack "regularExpression" .= (BL.unpack apdre)]
+    toJSON (Revocable r) = object [T.pack "revocable" .= r]
+    toJSON (KeyExpirationTime d) = object [T.pack "keyExpirationTime" .= d]
+    toJSON (PreferredSymmetricAlgorithms sas) = object [T.pack "preferredSymmetricAlgorithms" .= sas]
+    toJSON (RevocationKey rcs pka tof) = object [T.pack "revocationKey" .= (rcs, pka, tof)]
+    toJSON (Issuer eoki) = object [T.pack "issuer" .= eoki]
+    toJSON (NotationData nfs (NotationName nn) (NotationValue nv)) = object [T.pack "notationData" .= (nfs, BL.unpack nn, BL.unpack nv)]
+    toJSON (PreferredHashAlgorithms phas) = object [T.pack "preferredHashAlgorithms" .= phas]
+    toJSON (PreferredCompressionAlgorithms pcas) = object [T.pack "preferredCompressionAlgorithms" .=  pcas]
+    toJSON (KeyServerPreferences kspfs) = object [T.pack "keyServerPreferences" .= kspfs]
+    toJSON (PreferredKeyServer ks) = object [T.pack "preferredKeyServer" .= (show ks)]
+    toJSON (PrimaryUserId p) = object [T.pack "primaryUserId" .= p]
+    toJSON (PolicyURL u) = object [T.pack "policyURL" .= u]
+    toJSON (KeyFlags kfs) = object [T.pack "keyFlags" .= kfs]
+    toJSON (SignersUserId u) = object [T.pack "signersUserId" .= u]
+    toJSON (ReasonForRevocation rc rr) = object [T.pack "reasonForRevocation" .= (rc, rr)]
+    toJSON (Features ffs) = object [T.pack "features" .= ffs]
+    toJSON (SignatureTarget pka ha sh) = object [T.pack "signatureTarget" .= (pka, ha, BL.unpack sh)]
+    toJSON (EmbeddedSignature sp) = object [T.pack "embeddedSignature" .= sp]
+    toJSON (UserDefinedSigSub t bs) = object [T.pack "userDefinedSigSub" .= (t, BL.unpack bs)]
+    toJSON (OtherSigSub t bs) = object [T.pack "otherSigSub" .= (t, BL.unpack bs)]
+
+instance A.FromJSON SigSubPacketPayload where
+    parseJSON (A.Object v) = (SigCreationTime <$> v A..: T.pack "sigCreationTime")
+                         <|> (SigExpirationTime <$> v A..: T.pack "sigExpirationTime")
+                         <|> (ExportableCertification <$> v A..: T.pack "exportableCertification")
+                         <|> (uncurry TrustSignature <$> v A..: T.pack "trustSignature")
+                         <|> (RegularExpression . BL.pack <$> v A..: T.pack "regularExpression")
+                         <|> (Revocable <$> v A..: T.pack "revocable")
+                         <|> (KeyExpirationTime <$> v A..: T.pack "keyExpirationTime")
+                         <|> (PreferredSymmetricAlgorithms <$> v A..: T.pack "preferredSymmetricAlgorithms")
+                         <|> (uc3 RevocationKey <$> v A..: T.pack "revocationKey")
+                         <|> (Issuer <$> v A..: T.pack "issuer")
+                         <|> (uc3 NotationData <$> v A..: T.pack "notationData")
+    parseJSON _            = mzero
+
+uc3 :: (a -> b -> c -> d) -> ((a, b, c) -> d)
+uc3 f ~(a,b,c) = f a b c
 
 data HashAlgorithm = DeprecatedMD5
                    | SHA1
@@ -313,7 +368,7 @@ instance Eq HashAlgorithm where
     (==) a b = fromFVal a == fromFVal b
 
 instance Ord HashAlgorithm where
-    compare a b = fromFVal a `compare` fromFVal b
+    compare = comparing fromFVal
 
 instance FutureVal HashAlgorithm where
     fromFVal DeprecatedMD5 = 1
@@ -346,6 +401,7 @@ instance Pretty HashAlgorithm where
     pretty (OtherHA ha) = text "unknown hash algorithm type" <+> (text . show) ha
 
 instance A.ToJSON HashAlgorithm
+instance A.FromJSON HashAlgorithm
 
 data CompressionAlgorithm = Uncompressed
                           | ZIP
@@ -358,7 +414,7 @@ instance Eq CompressionAlgorithm where
     (==) a b = fromFVal a == fromFVal b
 
 instance Ord CompressionAlgorithm where
-    compare a b = fromFVal a `compare` fromFVal b
+    compare = comparing fromFVal
 
 instance FutureVal CompressionAlgorithm where
     fromFVal Uncompressed = 0
@@ -382,6 +438,7 @@ instance Pretty CompressionAlgorithm where
     pretty (OtherCA ca) = text "unknown compression algorithm type" <+> (text . show) ca
 
 instance A.ToJSON CompressionAlgorithm
+instance A.FromJSON CompressionAlgorithm
 
 class (Eq a, Ord a) => FutureVal a where
    fromFVal :: a -> Word8
@@ -403,7 +460,7 @@ instance Eq PubKeyAlgorithm where
     (==) a b = fromFVal a == fromFVal b
 
 instance Ord PubKeyAlgorithm where
-    compare a b = fromFVal a `compare` fromFVal b
+    compare = comparing fromFVal
 
 instance FutureVal PubKeyAlgorithm where
     fromFVal RSA = 1
@@ -442,6 +499,7 @@ instance Pretty PubKeyAlgorithm where
     pretty pka = text "unknown pubkey algorithm type" <+> (text . show) pka
 
 instance A.ToJSON PubKeyAlgorithm
+instance A.FromJSON PubKeyAlgorithm
 
 class (Eq a, Ord a) => FutureFlag a where
     fromFFlag :: a -> Int
@@ -455,7 +513,7 @@ instance Eq KSPFlag where
     (==) a b = fromFFlag a == fromFFlag b
 
 instance Ord KSPFlag where
-    compare a b = fromFFlag a `compare` fromFFlag b
+    compare = comparing fromFFlag
 
 instance FutureFlag KSPFlag where
     fromFFlag NoModify = 0
@@ -471,6 +529,7 @@ instance Pretty KSPFlag where
     pretty (KSPOther o) = text "unknown keyserver preference flag type" <+> pretty o
 
 instance A.ToJSON KSPFlag
+instance A.FromJSON KSPFlag
 
 data KeyFlag = GroupKey
              | AuthKey
@@ -486,7 +545,7 @@ instance Eq KeyFlag where
     (==) a b = fromFFlag a == fromFFlag b
 
 instance Ord KeyFlag where
-    compare a b = fromFFlag a `compare` fromFFlag b
+    compare = comparing fromFFlag
 
 instance FutureFlag KeyFlag where
     fromFFlag GroupKey = 0
@@ -520,6 +579,7 @@ instance Pretty KeyFlag where
     pretty (KFOther o) = text "unknown key flag type" <+> pretty o
 
 instance A.ToJSON KeyFlag
+instance A.FromJSON KeyFlag
 
 data RevocationClass = SensitiveRK
                      | RClOther Word8 -- FIXME: this should be constrained to 3 bits
@@ -529,7 +589,7 @@ instance Eq RevocationClass where
     (==) a b = fromFFlag a == fromFFlag b
 
 instance Ord RevocationClass where
-    compare a b = fromFFlag a `compare` fromFFlag b
+    compare = comparing fromFFlag
 
 instance FutureFlag RevocationClass where
     fromFFlag SensitiveRK = 1
@@ -545,6 +605,7 @@ instance Pretty RevocationClass where
     pretty (RClOther o) = text "unknown revocation class" <+> pretty o
 
 instance A.ToJSON RevocationClass
+instance A.FromJSON RevocationClass
 
 data RevocationCode = NoReason
                     | KeySuperseded
@@ -558,7 +619,7 @@ instance Eq RevocationCode where
     (==) a b = fromFVal a == fromFVal b
 
 instance Ord RevocationCode where
-    compare a b = fromFVal a `compare` fromFVal b
+    compare = comparing fromFVal
 
 instance FutureVal RevocationCode where
     fromFVal NoReason = 0
@@ -585,6 +646,7 @@ instance Pretty RevocationCode where
     pretty (RCoOther o) = text "unknown revocation code" <+> pretty o
 
 instance A.ToJSON RevocationCode
+instance A.FromJSON RevocationCode
 
 data FeatureFlag = ModificationDetection
                  | FeatureOther Int
@@ -594,7 +656,7 @@ instance Eq FeatureFlag where
     (==) a b = fromFFlag a == fromFFlag b
 
 instance Ord FeatureFlag where
-    compare a b = fromFFlag a `compare` fromFFlag b
+    compare = comparing fromFFlag
 
 instance FutureFlag FeatureFlag where
     fromFFlag ModificationDetection = 7
@@ -612,6 +674,7 @@ instance Pretty FeatureFlag where
     pretty (FeatureOther o) = text "unknown feature flag type" <+> pretty o
 
 instance A.ToJSON FeatureFlag
+instance A.FromJSON FeatureFlag
 
 newtype MPI = MPI {unMPI :: Integer}
     deriving (Data, Eq, Generic, Show, Typeable)
@@ -626,6 +689,7 @@ instance Pretty MPI where
     pretty = pretty . unpack
 
 instance A.ToJSON MPI
+instance A.FromJSON MPI
 
 data SignaturePayload = SigV3 SigType ThirtyTwoBitTimeStamp EightOctetKeyId PubKeyAlgorithm HashAlgorithm Word16 (NonEmpty MPI)
                       | SigV4 SigType PubKeyAlgorithm HashAlgorithm [SigSubPacket] [SigSubPacket] Word16 (NonEmpty MPI)
@@ -654,6 +718,7 @@ instance Pretty KeyVersion where
     pretty V4 = text "v4"
 
 instance A.ToJSON KeyVersion
+instance A.FromJSON KeyVersion
 
 data PKPayload = PKPayload {
       _keyVersion :: KeyVersion
@@ -664,7 +729,7 @@ data PKPayload = PKPayload {
     } deriving (Data, Eq, Generic, Show, Typeable)
 
 instance Ord PKPayload where
-    compare = comparing _keyVersion <> comparing _timestamp <> comparing _v3exp <> comparing _pkalgo <> comparing (show . _pubkey) -- FIXME: this is suboptimal
+    compare = comparing _keyVersion <> comparing _timestamp <> comparing _v3exp <> comparing _pkalgo <> comparing _pubkey
 
 instance Hashable PKPayload
 
@@ -721,7 +786,7 @@ instance Eq DataType where
     (==) a b = fromFVal a == fromFVal b
 
 instance Ord DataType where
-    compare a b = fromFVal a `compare` fromFVal b
+    compare = comparing fromFVal
 
 instance FutureVal DataType where
     fromFVal BinaryData = fromIntegral . fromEnum $ 'b'
@@ -741,6 +806,7 @@ instance Pretty DataType where
     pretty (OtherData o) = text "other data type " <+> (text . show) o
 
 instance A.ToJSON DataType
+instance A.FromJSON DataType
 
 newtype Salt = Salt {unSalt :: B.ByteString}
     deriving (Byteable, Data, Eq, Generic, Hashable, Show, Typeable)
@@ -766,6 +832,7 @@ instance Pretty IterationCount where
     pretty = pretty . unpack
 
 instance A.ToJSON IterationCount
+instance A.FromJSON IterationCount
 
 data S2K = Simple HashAlgorithm
          | Salted HashAlgorithm Salt
@@ -819,6 +886,7 @@ instance Pretty ImageHeader where
     pretty (ImageHV1 f) = text "imghdr v1" <+> pretty f
 
 instance A.ToJSON ImageHeader
+instance A.FromJSON ImageHeader
 
 data ImageFormat = JPEG
                  | OtherImage Word8
@@ -828,7 +896,7 @@ instance Eq ImageFormat where
     (==) a b = fromFVal a == fromFVal b
 
 instance Ord ImageFormat where
-    compare a b = fromFVal a `compare` fromFVal b
+    compare = comparing fromFVal
 
 instance FutureVal ImageFormat where
     fromFVal JPEG = 1
@@ -844,6 +912,7 @@ instance Pretty ImageFormat where
     pretty (OtherImage o) = text "unknown image format" <+> pretty o
 
 instance A.ToJSON ImageFormat
+instance A.FromJSON ImageFormat
 
 data SigType = BinarySig
              | CanonicalTextSig
@@ -867,7 +936,7 @@ instance Eq SigType where
     (==) a b = fromFVal a == fromFVal b
 
 instance Ord SigType where
-    compare a b = fromFVal a `compare` fromFVal b
+    compare = comparing fromFVal
 
 instance FutureVal SigType where
     fromFVal BinarySig = 0x00
@@ -925,6 +994,7 @@ instance Pretty SigType where
     pretty (OtherSig o) = text "unknown signature type" <+> pretty o
 
 instance A.ToJSON SigType
+instance A.FromJSON SigType
 
 newtype DSA_PublicKey = DSA_PublicKey {unDSA_PublicKey :: DSA.PublicKey}
     deriving (Data, Eq, Generic, Show, Typeable)
@@ -1069,7 +1139,12 @@ instance Read EightOctetKeyId where
 instance Hashable EightOctetKeyId
 
 instance A.ToJSON EightOctetKeyId where
-    toJSON = A.toJSON . w8sToHex . BL.unpack . unpack
+    toJSON e = object [T.pack "eoki" .= (w8sToHex . BL.unpack . unpack) e]
+
+instance A.FromJSON EightOctetKeyId where
+    parseJSON (A.Object v) = EightOctetKeyId . read <$>
+                                      v A..: T.pack "eoki"
+    parseJSON _            = mzero
 
 newtype TwentyOctetFingerprint = TwentyOctetFingerprint {unTOF :: ByteString}
     deriving (Data, Eq, Generic, Ord, Show, Typeable)
@@ -1088,7 +1163,11 @@ instance Pretty TwentyOctetFingerprint where
     pretty = text . take 40 . w8sToHex . BL.unpack . unTOF
 
 instance A.ToJSON TwentyOctetFingerprint where
-    toJSON = A.toJSON . show . pretty
+    toJSON e = object [T.pack "fpr" .= (A.toJSON . show . pretty) e]
+instance A.FromJSON TwentyOctetFingerprint where
+    parseJSON (A.Object v) = TwentyOctetFingerprint . read <$>
+                                      v A..: T.pack "fpr"
+    parseJSON _            = mzero
 
 newtype SpacedFingerprint = SpacedFingerprint { unSpacedFingerprint :: TwentyOctetFingerprint }
 
@@ -1111,11 +1190,13 @@ data TK = TK {
   , _tkUIDs :: [(Text, [SignaturePayload])]
   , _tkUAts :: [([UserAttrSubPacket], [SignaturePayload])]
   , _tkSubs :: [(Pkt, [SignaturePayload])]
-  } deriving (Data, Eq, Show, Typeable)
+  } deriving (Data, Eq, Generic, Show, Typeable)
 
 instance Ord TK where
     compare = comparing _tkKey -- FIXME: is this ridiculous?
 
+instance A.ToJSON TK
+
 type Keyring = IxSet TK
 
 class Packet a where
@@ -1218,7 +1299,7 @@ data PKESK = PKESK
     { _pkeskPacketVersion :: PacketVersion
     , _pkeskEightOctetKeyId :: EightOctetKeyId
     , _pkeskPubKeyAlgorithm :: PubKeyAlgorithm
-    , _pkeskMPIs :: (NonEmpty MPI)
+    , _pkeskMPIs :: NonEmpty MPI
     } deriving (Data, Eq, Show, Typeable)
 instance Packet PKESK where
     data PacketType PKESK = PKESKType deriving (Show, Eq)
@@ -1226,6 +1307,7 @@ instance Packet PKESK where
     packetCode _ = 1
     toPkt (PKESK a b c d) = PKESKPkt a b c d
     fromPkt (PKESKPkt a b c d) = PKESK a b c d
+    fromPkt _ = error "Cannot coerce non-PKESK packet"
 
 instance Pretty PKESK where
     pretty = pretty . toPkt
@@ -1237,8 +1319,9 @@ instance Packet Signature where
     data PacketType Signature = SignatureType deriving (Show, Eq)
     packetType _ = SignatureType
     packetCode _ = 2
-    toPkt (Signature a ) = SignaturePkt a
+    toPkt (Signature a) = SignaturePkt a
     fromPkt (SignaturePkt a) = Signature a
+    fromPkt _ = error "Cannot coerce non-Signature packet"
 
 instance Pretty Signature where
     pretty = pretty . toPkt
@@ -1255,6 +1338,7 @@ instance Packet SKESK where
     packetCode _ = 3
     toPkt (SKESK a b c d) = SKESKPkt a b c d
     fromPkt (SKESKPkt a b c d) = SKESK a b c d
+    fromPkt _ = error "Cannot coerce non-SKESK packet"
 
 instance Pretty SKESK where
     pretty = pretty . toPkt
@@ -1273,6 +1357,7 @@ instance Packet OnePassSignature where
     packetCode _ = 4
     toPkt (OnePassSignature a b c d e f) = OnePassSignaturePkt a b c d e f
     fromPkt (OnePassSignaturePkt a b c d e f) = OnePassSignature a b c d e f
+    fromPkt _ = error "Cannot coerce non-OnePassSignature packet"
 
 instance Pretty OnePassSignature where
     pretty = pretty . toPkt
@@ -1287,6 +1372,7 @@ instance Packet SecretKey where
     packetCode _ = 5
     toPkt (SecretKey a b) = SecretKeyPkt a b
     fromPkt (SecretKeyPkt a b) = SecretKey a b
+    fromPkt _ = error "Cannot coerce non-SecretKey packet"
 
 instance Pretty SecretKey where
     pretty = pretty . toPkt
@@ -1300,6 +1386,7 @@ instance Packet PublicKey where
     packetCode _ = 6
     toPkt (PublicKey a) = PublicKeyPkt a
     fromPkt (PublicKeyPkt a) = PublicKey a
+    fromPkt _ = error "Cannot coerce non-PublicKey packet"
 
 instance Pretty PublicKey where
     pretty = pretty . toPkt
@@ -1314,6 +1401,7 @@ instance Packet SecretSubkey where
     packetCode _ = 7
     toPkt (SecretSubkey a b) = SecretSubkeyPkt a b
     fromPkt (SecretSubkeyPkt a b) = SecretSubkey a b
+    fromPkt _ = error "Cannot coerce non-SecretSubkey packet"
 
 instance Pretty SecretSubkey where
     pretty = pretty . toPkt
@@ -1328,6 +1416,7 @@ instance Packet CompressedData where
     packetCode _ = 8
     toPkt (CompressedData a b) = CompressedDataPkt a b
     fromPkt (CompressedDataPkt a b) = CompressedData a b
+    fromPkt _ = error "Cannot coerce non-CompressedData packet"
 
 instance Pretty CompressedData where
     pretty = pretty . toPkt
@@ -1341,6 +1430,7 @@ instance Packet SymEncData where
     packetCode _ = 9
     toPkt (SymEncData a) = SymEncDataPkt a
     fromPkt (SymEncDataPkt a) = SymEncData a
+    fromPkt _ = error "Cannot coerce non-SymEncData packet"
 
 instance Pretty SymEncData where
     pretty = pretty . toPkt
@@ -1354,6 +1444,7 @@ instance Packet Marker where
     packetCode _ = 10
     toPkt (Marker a) = MarkerPkt a
     fromPkt (MarkerPkt a) = Marker a
+    fromPkt _ = error "Cannot coerce non-Marker packet"
 
 instance Pretty Marker where
     pretty = pretty . toPkt
@@ -1370,6 +1461,7 @@ instance Packet LiteralData where
     packetCode _ = 11
     toPkt (LiteralData a b c d) = LiteralDataPkt a b c d
     fromPkt (LiteralDataPkt a b c d) = LiteralData a b c d
+    fromPkt _ = error "Cannot coerce non-LiteralData packet"
 
 instance Pretty LiteralData where
     pretty = pretty . toPkt
@@ -1383,6 +1475,7 @@ instance Packet Trust where
     packetCode _ = 12
     toPkt (Trust a) = TrustPkt a
     fromPkt (TrustPkt a) = Trust a
+    fromPkt _ = error "Cannot coerce non-Trust packet"
 
 instance Pretty Trust where
     pretty = pretty . toPkt
@@ -1396,6 +1489,7 @@ instance Packet UserId where
     packetCode _ = 13
     toPkt (UserId a) = UserIdPkt a
     fromPkt (UserIdPkt a) = UserId a
+    fromPkt _ = error "Cannot coerce non-UserId packet"
 
 instance Pretty UserId where
     pretty = pretty . toPkt
@@ -1409,6 +1503,7 @@ instance Packet PublicSubkey where
     packetCode _ = 14
     toPkt (PublicSubkey a) = PublicSubkeyPkt a
     fromPkt (PublicSubkeyPkt a) = PublicSubkey a
+    fromPkt _ = error "Cannot coerce non-PublicSubkey packet"
 
 instance Pretty PublicSubkey where
     pretty = pretty . toPkt
@@ -1422,6 +1517,7 @@ instance Packet UserAttribute where
     packetCode _ = 17
     toPkt (UserAttribute a) = UserAttributePkt a
     fromPkt (UserAttributePkt a) = UserAttribute a
+    fromPkt _ = error "Cannot coerce non-UserAttribute packet"
 
 instance Pretty UserAttribute where
     pretty = pretty . toPkt
@@ -1436,6 +1532,7 @@ instance Packet SymEncIntegrityProtectedData where
     packetCode _ = 18
     toPkt (SymEncIntegrityProtectedData a b) = SymEncIntegrityProtectedDataPkt a b
     fromPkt (SymEncIntegrityProtectedDataPkt a b) = SymEncIntegrityProtectedData a b
+    fromPkt _ = error "Cannot coerce non-SymEncIntegrityProtectedData packet"
 
 instance Pretty SymEncIntegrityProtectedData where
     pretty = pretty . toPkt
@@ -1449,6 +1546,7 @@ instance Packet ModificationDetectionCode where
     packetCode _ = 19
     toPkt (ModificationDetectionCode a) = ModificationDetectionCodePkt a
     fromPkt (ModificationDetectionCodePkt a) = ModificationDetectionCode a
+    fromPkt _ = error "Cannot coerce non-ModificationDetectionCode packet"
 
 instance Pretty ModificationDetectionCode where
     pretty = pretty . toPkt
@@ -1463,6 +1561,7 @@ instance Packet OtherPacket where
     packetCode _ = undefined -- FIXME
     toPkt (OtherPacket a b) = OtherPacketPkt a b
     fromPkt (OtherPacketPkt a b) = OtherPacket a b
+    fromPkt _ = error "Cannot coerce non-OtherPacket packet"
 
 instance Pretty OtherPacket where
     pretty = pretty . toPkt
@@ -1478,6 +1577,7 @@ instance Packet BrokenPacket where
     packetCode _ = undefined
     toPkt (BrokenPacket a b c) = BrokenPacketPkt a b c
     fromPkt (BrokenPacketPkt a b c) = BrokenPacket a b c
+    fromPkt _ = error "Cannot coerce non-BrokenPacket packet"
 
 instance Pretty BrokenPacket where
     pretty = pretty . toPkt
